% Shrutilipi Bhattacharjee
% Basic Semantic Kriging (OK) with Motivating Example
% 20190505
% Input parameters: 
% GeoTiff of Primary Parameter (with missing value as NaN)
% GeoTiff of LULC (with colourCode)
%     Cropland (C)              1 (Row/Colum of SS/SI matrix)
%     NaturalVegetation (NV)    2 (Row/Colum of SS/SI matrix)
%     UrbanArea (UA)            3 (Row/Colum of SS/SI matrix)
%     BareArea (BA)             4 (Row/Colum of SS/SI matrix)
%     Waterbodies (W)           5 (Row/Colum of SS/SI matrix)
% Constraints: 
% Maximum and minimum number of interpolating points = 24 and 2
% Number of LULC classes = 5

clear all;
close all;

% Input the raster with missing values as NaN
[PrimaryParameter, R1] = geotiffread('MotivatingExample_XCO2');  

% Determine the number of rows and columns in the raster
[Row,Col] = size(PrimaryParameter);

% Input the maximum number of interpolating points
InterpolatingPoints = 10; % Default number of maximum interpolating points

% Change the maximum number of interpolating points
prompt = 'Please enter the maximum number of interpolating ponits:::';
InterpolatingPoints = input(prompt);

%%%%%%%%%%%%%%%%%%%%%%% SemK Processing %%%%%%%%%%%%%%%%%%%%%%%
% Define the ontology and enter the path of every LULC class as
% a row. Declare every LULC as maximum hop, filling rest of the
% hops as empty string ('')
Ontology(1,:) = {'LULC', 'Cropland'};
Ontology(2,:) = {'LULC', 'NaturalVegetation'};
Ontology(3,:) = {'LULC', 'UrbanArea'};
Ontology(4,:) = {'LULC', 'BareArea'};
Ontology(5,:) = {'LULC', 'Waterbodies'};
            
% Determine the number of LULC classes
LULCclasses = size(Ontology,1); % Change as per scenario
            
% Determine the semantic similarity (SS) matrix: SS
SS = NaN(LULCclasses,LULCclasses);
for m=1:1:LULCclasses
    for n=m:1:LULCclasses
        Counter = 0;
        mi = 0;
        while Counter<size(Ontology(m,:),2) && Counter<size(Ontology(n,:),2)
            if strcmp(Ontology(m,Counter+1),Ontology(n,Counter+1))
                mi = mi + 1;
            else
                break;
            end
            Counter = Counter + 1;
        end
        SS(m,n) = (mi/(size(Ontology(m,:),2)) + mi/(size(Ontology(n,:),2)))/2;
        SS(n,m) = (mi/(size(Ontology(m,:),2)) + mi/(size(Ontology(n,:),2)))/2;
    end
end

% Determine the temporary spatial importance (SI) matrix: SI_Prev
[LULC, R2] = geotiffread('MotivatingExample_LULC'); 
SI_Prev = ones(LULCclasses,LULCclasses);
Correlation = NaN(size(LULC,2)*size(LULC,2),2);
LULC1 = 1;
while (LULC1<=5)
    LULC2 = LULC1 + 1;
    while (LULC2<=5)
        Counter = 0;
        for p = 1:1:size(LULC,2)
            for q = 1:1:size(LULC,1)
                if LULC(p,q) == LULC1
                    for m = -1:1:1
                        for n = -1:1:1
                            if (p+m)>0 && (p+m)<=size(LULC,2) && (q+n)>0 && (q+n)<=size(LULC,1) && LULC(p+m,q+n) == LULC2
                                Correlation(Counter+1,1) = PrimaryParameter(p,q);
                                Correlation(Counter+1,2) = PrimaryParameter(p+m,q+n);
                                Counter = Counter + 1;
                            end
                        end
                    end
                end
            end                      
        end
        Correlation = Correlation(all(~isnan(Correlation),2),:); 
        SI_Prev(LULC1,LULC2) = corr(Correlation(:,1),Correlation(:,2));
        SI_Prev(LULC2,LULC1) = corr(Correlation(:,1),Correlation(:,2));
        LULC2 = LULC2 + 1; 
    end                
    LULC1 = LULC1 + 1;
end

% Determine the spatial importance (SI) matrix: SI, by normalizing the
% SI_Prev matrix in the range of (0.0001,1)
SI = rescale(SI_Prev,0.0001,1);
         
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
            
% Calculate the prediction zone, its size parameters
Max = Row;
Corner = (Row+1)/2;
for i = 1:Row
    for j = 1:Col
        if isnan(PrimaryParameter(i,j))
            fprintf('Prediction location is (%d, %d)',[i,j]);
                         
            %Creation of prediction zone: PredictionZone
            PredictionZoneParameter = NaN(Max,Max);
            PredictionZoneLULC = NaN(Max,Max);
            Flag = zeros(Max,Max);
            Counter = 0;
            Max_temp = 1;
            Corner_temp = 0;
            Mid = (Max+1)/2;
            while Max_temp <= Max && Corner_temp <= Corner 
                for m=(-1*Corner_temp):1:(1*Corner_temp)
                    for n=(-1*Corner_temp):1:(1*Corner_temp)
                        if Counter<=InterpolatingPoints && (i+m)>0 && (j+n)>0 && (i+m)<=Row && (j+n)<=Col && Flag(Mid+m,Mid+n)==0
                            PredictionZoneParameter(Mid+m,Mid+n) = PrimaryParameter((i+m),(j+n));
                          	PredictionZoneLULC(Mid+m,Mid+n) = LULC((i+m),(j+n));
                            Flag(Mid+m,Mid+n) = 1;
                            Counter = Counter + 1;
                       	end
                        if m == 0 && n == 0
                            PredictionZoneParameter(Mid+m,Mid+n) = 1000000;
                        end
                    end  
               	end
                Max_temp = Max_temp + 2;
                Corner_temp = Corner_temp + 1;
            end
                      
            PredictionZoneParameter( ~any(PredictionZoneParameter,2), : ) = []; 
            PredictionZoneParameter( :, ~any(PredictionZoneParameter,1) ) = [];
            PredictionZoneLULC( ~any(PredictionZoneLULC,2), : ) = []; 
            PredictionZoneLULC( :, ~any(PredictionZoneLULC,1) ) = [];
            
            
            % Flag the prediction point in the prediction zone
            % Determine the Z(xi): ValInterpolatingPoints
            PredLocRow = 0;
            PredLocCol = 0;
            ValInterpolatingPoints = NaN(Max*Max,1);
            Counter = 0;
            
            for m = 1:1:size(PredictionZoneParameter,1)
                for n = 1:1:size(PredictionZoneParameter,2)
                	if PredictionZoneParameter(m,n) == 1000000
                        PredLocRow = m;
                        PredLocCol = n;
                        PredictionZoneParameter(m,n) = NaN;
                    end
                    ValInterpolatingPoints(Counter+1,1) = PredictionZoneParameter(m,n);
                    Counter = Counter +1;
                end
            end
                   
            ValInterpolatingPoints(ValInterpolatingPoints == 0) = [];
            ValInterpolatingPoints = ValInterpolatingPoints(~isnan(ValInterpolatingPoints));
            NonNaN = nnz(~isnan(PredictionZoneParameter));
            
            % Determine the pairwise Euclidean distance matrices: SemivarianceDist(NxN), DistanceDist(Nx1)
            % Determine the pairwise temporary semivariance matrix: Semivariance_Prev(NxN)
            SemivarianceDist = NaN(size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1),size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1));  
            DistanceDist = zeros(size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1),size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1));  
            SemivarianceSem = NaN(size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1),size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1));  
            DistanceSem = NaN(size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1),size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1));  
            Semivariance_Prev = NaN(size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1),size(PredictionZoneParameter,2)*size(PredictionZoneParameter,1));      
            for p = 1:1:size(PredictionZoneParameter,1)
                for q = 1:1:size(PredictionZoneParameter,2)
                    for m = 1:1:size(PredictionZoneParameter,1)
                        for n = 1:1:size(PredictionZoneParameter,2)
                            if ~isnan(PredictionZoneParameter(p,q)) && ~isnan(PredictionZoneParameter(m,n)) 
                                SemivarianceDist((p-1)*size(PredictionZoneParameter,2)+q,(m-1)*size(PredictionZoneParameter,2)+n) = sqrt((p-m)^2+(q-n)^2);
                                SemivarianceSem((p-1)*size(PredictionZoneParameter,2)+q,(m-1)*size(PredictionZoneParameter,2)+n) = SS(PredictionZoneLULC(p,q),PredictionZoneLULC(m,n))*SI(PredictionZoneLULC(p,q),PredictionZoneLULC(m,n));
                                Semivariance_Prev((p-1)*size(PredictionZoneParameter,2)+q,(m-1)*size(PredictionZoneParameter,2)+n) = (PredictionZoneParameter(p,q) - PredictionZoneParameter(m,n)).^2;
                            elseif p==PredLocRow && q==PredLocCol && ~isnan(PredictionZoneParameter(m,n))
                                DistanceDist((p-1)*size(PredictionZoneParameter,2)+q,(m-1)*size(PredictionZoneParameter,2)+n) = sqrt((p-m)^2+(q-n)^2);
                                DistanceSem((p-1)*size(PredictionZoneParameter,2)+q,(m-1)*size(PredictionZoneParameter,2)+n) = SS(PredictionZoneLULC(p,q),PredictionZoneLULC(m,n))*SI(PredictionZoneLULC(p,q),PredictionZoneLULC(m,n));
                            elseif ~isnan(PredictionZoneParameter(p,q)) && m==PredLocRow && n==PredLocCol  
                                DistanceDist((p-1)*size(PredictionZoneParameter,2)+q,(m-1)*size(PredictionZoneParameter,2)+n) = sqrt((p-m)^2+(q-n)^2);
                            else 
                                DistanceDist((p-1)*size(PredictionZoneParameter,2)+q,(m-1)*size(PredictionZoneParameter,2)+n) = 0;
                            end
                        end
                    end
                end    
            end
            
            Semivariance_Prev(~any(Semivariance_Prev,2),:) = []; 
            Semivariance_Prev(:,~any(Semivariance_Prev,1)) = [];
            Semivariance_Prev = Semivariance_Prev(all(~isnan(Semivariance_Prev),2),:); 
            Semivariance_Prev = Semivariance_Prev(:,all(~isnan(Semivariance_Prev)));

            SemivarianceDist(~any(SemivarianceDist,2),:) = []; 
            SemivarianceDist(:,~any(SemivarianceDist,1)) = [];
            SemivarianceDist = SemivarianceDist(all(~isnan(SemivarianceDist),2),:); 
            SemivarianceDist = SemivarianceDist(:,all(~isnan(SemivarianceDist)));
            
            SemivarianceSem(~any(SemivarianceSem,2),:) = []; 
            SemivarianceSem(:,~any(SemivarianceSem,1)) = [];
            SemivarianceSem = SemivarianceSem(all(~isnan(SemivarianceSem),2),:); 
            SemivarianceSem = SemivarianceSem(:,all(~isnan(SemivarianceSem)));
                       
         	DistanceDist = DistanceDist(:,((PredLocRow-1)*size(PredictionZoneParameter,2)+PredLocCol));
            DistanceDist = DistanceDist(DistanceDist~= 0);
            
            DistanceSem = DistanceSem(((PredLocRow-1)*size(PredictionZoneParameter,2)+PredLocCol),:);
            DistanceSem = DistanceSem(~isnan(DistanceSem))';
            
            % Determine the pairwise semivariance matrix: Semivariance(NxN)
            Semivariance = NaN;
            
            % Determine the average semivariance score at every unique distance
            Count = zeros(size(unique(SemivarianceDist),1),1);
            SumSemivariance = zeros(size(unique(SemivarianceDist),1),1);
            UniqueDist = unique(SemivarianceDist);
            Counter = 1;
            while Counter <= size(UniqueDist,1)
                for p = 1:1:size(Semivariance_Prev,2)
                    for q = 1:1:size(Semivariance_Prev,1)
                        if SemivarianceDist(p,q) == UniqueDist(Counter,1)
                            Count(Counter) = Count(Counter)+1;
                            SumSemivariance(Counter) = SumSemivariance(Counter) + Semivariance_Prev(p,q);                   
                        end
                  	end
                end
                Counter = Counter + 1;
            end                            
            
            % Assign the average semivariance score to every unique distance
            for p = 1:1:size(SemivarianceDist)
                for q = 1:1:size(SemivarianceDist)
                    Counter = 1;
                    while Counter <= size(UniqueDist,1)
                        if SemivarianceDist(p,q) == UniqueDist(Counter,1)
                            Semivariance(p,q) = SumSemivariance(Counter)/(2*Count(Counter));
                        end
                        Counter = Counter + 1;
                    end
                end
            end
            
            %Plot the experimental semivariogram
           	scatter(SemivarianceDist(:),Semivariance(:),'filled')
            ft=fittype('exp1');
            ExpSemivariogram=fit(SemivarianceDist(:),Semivariance(:),ft);
            ExpSemivariogram;
            hold on
            plot(ExpSemivariogram);
            hold off
            xlabel('Distance (h)');
            ylabel('Semivariance (\gamma)');
            title('Experimental semivariogram (Exponential model)');
            Sill = feval(ExpSemivariogram, max(SemivarianceDist(:)));
            Nugget = feval(ExpSemivariogram, 0);
            PartialSill = Sill - Nugget;
            
            % Determine the covariance matrix (C): Covariance(NxN) with
            % Exponential model
            Covariance = NaN;
            for p = 1:1:size(SemivarianceDist,2)
                for q = 1:1:size(SemivarianceDist,1)
                    if SemivarianceDist(p,q) == 0
                        Covariance(p,q) = 0;
                    else
                        Covariance(p,q) = (Sill - Nugget * (1 - exp(-SemivarianceDist(p,q)/(max(SemivarianceDist(:))/3))))*(SemivarianceSem(p,q));
                    end
                end
            end
            
            % Determine the distance matrix (D): Distance(Nx1)
            Distance = NaN;
            for p = 1:1:size(DistanceDist,1)
                Distance(p,1) = (Sill - Nugget * (1 - exp(-DistanceDist(p)/(max(SemivarianceDist(:))/3))))*(DistanceSem(p,1));
            end
            
            % Determine the weight vector of OK: W_OK(Nx1)
            PrevW_SemK = NaN;
            W_SemK = NaN;
            PrevW_SemK = Covariance\Distance;
            for p = 1:1:size(PrevW_SemK,1)
                W_SemK(p,1)=(PrevW_SemK(p)- min(PrevW_SemK))/(max(PrevW_SemK)-min(PrevW_SemK)) * (3-1)+1;
            end
            W_SemK = normalize(W_SemK,'norm',1);
            
            % Calculate the missing value: PrimaryParameter(i,j)
            PrimaryParameter(i,j) = dot(W_SemK,ValInterpolatingPoints);
            fprintf(' and the predicted value is: %d \n', PrimaryParameter(i,j));
        end
    end
end